/*
 DSPAM
 COPYRIGHT (C) 2004 NETWORK DWEEBS CORPORATION

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifdef HAVE_CONFIG_H
#include <auto-config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/param.h>
#include "config.h"
#include "util.h"

#ifdef TIME_WITH_SYS_TIME
#   include <sys/time.h>
#   include <time.h>
#else
#   ifdef HAVE_SYS_TIME_H
#       include <sys/time.h>
#   else
#       include <time.h>
#   endif
#endif

#include "dspam.h"
#include "libdspam.h"
#include "language.h"
#include "buffer.h"
#include "base64.h"
#include "md5.h"
#include "tbt.h"

#ifdef SYSTEM_LOGGING
  struct timeval tv_start, tv_end;
#endif

int
main (int argc, char *argv[])
{
  AGENT_CTX ATX;		/* Our agent context */

  struct nt *users;             /* Destination users */
  struct nt_node *node_nt;
  struct nt_c c_nt;

  buffer *message;              /* Message */
  char buff[1024];              /* Read Buffer */

#ifdef DEBUG
  char debug_args[1024];        /* Report on all arguments passed in */
#endif

  int i, line = 1;
  int user_flag = 0;		/* Next argument(s) are users */
#ifdef TRUSTED_USER_SECURITY
  int trusted = 0;
#endif

  struct passwd *p;
  int retcode, exitcode = EXIT_SUCCESS;

#ifdef SYSTEM_LOGGING
  gettimeofday(&tv_start, NULL);
#endif

  p = getpwuid (getuid ());
  srand (getpid ());

  umask (006);                  /* rw-rw---- */
  setbuf (stdout, NULL);

  users   = nt_create (NT_CHAR);
  message = buffer_create (NULL);

  if (users == NULL || message == NULL)
  {
    report_error (ERROR_MEM_ALLOC);
    exit (EXIT_FAILURE);
  }

  /* Initialize Agent Context */
  memset(&ATX, 0, sizeof(AGENT_CTX));
  ATX.operating_mode  = -1;
  ATX.training_mode   = -1;
  ATX.training_buffer = 5;
  ATX.classification  = -1;
  ATX.source          = -1;
  ATX.operating_mode  = DSM_PROCESS;

#ifdef TRUSTED_USER_SECURITY
  trusted = is_trusted(p);
  if (!trusted)
  {
    nt_add (users, p->pw_name);
  }
#endif

  snprintf(ATX.mailer_args, sizeof(ATX.mailer_args), "%s ", MLOCAL);

#ifdef DEBUG
  debug_args[0] = 0;
#endif

  /* Process Commandline Arguments */

  for (i = 0; i < argc; i++)
  {

#ifdef DEBUG
    strlcat (debug_args, argv[i], sizeof (debug_args));
    strlcat (debug_args, " ", sizeof (debug_args));
#endif

    if (user_flag &&
       (argv[i][0] == '-' || argv[i][0] == 0 || !strcmp(argv[i], "--")))
    {
       user_flag = 0;
       if (!strcmp(argv[i], "--"))
         continue;
    }

    /* Set training mode */
    if (!strncmp (argv[i], "--mode=", 7))
    {
      char *mode = strchr(argv[i], '=')+1;
      if (!strcmp(mode, "toe"))
        ATX.training_mode = DST_TOE;
      else if (!strcmp(mode, "teft"))
        ATX.training_mode = DST_TEFT;
      else if (!strcmp(mode, "tum"))
        ATX.training_mode = DST_TUM;
      else if (!strcmp(mode, "notrain"))
        ATX.training_mode = DST_NOTRAIN;
      else
      {
        report_error_printf("invalid mode '%s'", mode);
        exit(EXIT_FAILURE);
      }
      
      continue;
    }

    /* Set runtime target user(s) */

    if (!strcmp (argv[i], "--user"))
    {
#ifdef TRUSTED_USER_SECURITY
      if (!trusted) {
        report_error(ERROR_UNTRUSTED_SET);
        exit(EXIT_FAILURE);
      }
#endif
      user_flag = 1;
      continue;
    }

    if (user_flag)
    {
      if (argv[i] != NULL && strlen (argv[i]) <= MAX_USERNAME_LENGTH)
      {
        nt_add (users, argv[i]);
      }
    }

    /* Print Syntax and Exit */

    if (!strcmp (argv[i], "--help"))
    {
      fprintf (stderr, "%s\n", SYNTAX);
      exit(EXIT_SUCCESS);
    }

    /* Print Version and Exit */

    if (!strcmp (argv[i], "--version"))
    {
      printf ("\nDSPAM Anti-Spam Suite %s (agent/library)\n\n", VERSION);
      printf ("Copyright (c) 2003 Network Dweebs Corporation\n");
      printf ("http://www.nuclearelephant.com/projects/dspam/\n\n");
      printf ("DSPAM may be copied only under the terms of the GNU "
              "General Public License,\n");
      printf ("a copy of which can be found with the DSPAM distribution "
              "kit.\n\n");
      exit (EXIT_SUCCESS);
    }

    /* If the message already has a classification */

    if (!strncmp (argv[i], "--class=", 8))
    {
      char *c = strchr(argv[i], '=')+1;
      if (!strcmp(c, "spam"))
        ATX.classification = DSR_ISSPAM;
      else if (!strcmp(c, "innocent"))
        ATX.classification = DSR_ISINNOCENT;
      else
      {
        report_error_printf("unknown classification '%s'", c);
        exit(EXIT_FAILURE);
      }

      continue;
    }

    /*
       The source of the classification:
         error: classification error made by dspam
         corpus: message from user's corpus
         inoculation: message inoculation
    */

    if (!strncmp (argv[i], "--source=", 9))
    {
      char *s = strchr(argv[i], '=')+1;

      if (!strcmp(s, "corpus"))
        ATX.source = DSS_CORPUS;
      else if (!strcmp(s, "inoculation"))
        ATX.source = DSS_INOCULATION;
      else if (!strcmp(s, "error"))
        ATX.source = DSS_ERROR;
      else
      {
        report_error_printf("unknown source '%s'", s); 
        exit(EXIT_FAILURE);
      }
      continue;
    }

    /* Operating Mode: Classify Only. */
    if (!strcmp (argv[i], "--classify"))
    {
      ATX.operating_mode = DSM_CLASSIFY;
      ATX.training_mode = DST_NOTRAIN;
      continue;
    }

    /* Operating Mode: Process Message */
    if (!strcmp (argv[i], "--process"))
    {
      ATX.operating_mode = DSM_PROCESS;
      continue;
    }

    /*
      Which messages should be delivered? 
      spam,innocent
      If spam is not specified, standard quarantine procedure will be used
    */

    if (!strncmp (argv[i], "--deliver=", 10))
    {
      char *d = strdup(strchr(argv[i], '=')+1);
      char *s;
      if (d == NULL) {
        report_error(ERROR_MEM_ALLOC);
        exit(EXIT_FAILURE);
      }

      s = strtok(d, ",");
      while(s != NULL) {
        if (!strcmp(s, "spam")) 
          ATX.flags |= DAF_DELIVER_SPAM;
        else if (!strcmp(s, "innocent"))
          ATX.flags |= DAF_DELIVER_INNOCENT;
        else
        {
          report_error_printf("unknown deliver type '%s'", s);
          exit(EXIT_FAILURE);
        }
      
        s = strtok(NULL, ",");
      }
      free(d);
      continue;
    }


    /* 
      Which features should be enabled?
      chained,noise,whitelist

      chained:   chained tokens (nGrams)
      noise:     bayesian noise reduction
      whitelist: automatic whitelisting
      tb=N:      set training buffer level (0-10)

      all features have their own internal instantiation requirements
    */

    if (!strncmp (argv[i], "--feature=", 10))
    {
      char *d = strdup(strchr(argv[i], '=')+1);
      char *s;
      if (d == NULL) {
        report_error(ERROR_MEM_ALLOC);
        exit(EXIT_FAILURE);
      }
                                                                                
      s = strtok(d, ",");
      while(s != NULL) {
        if (!strcmp(s, "chained") || !strcmp(s, "ch"))
          ATX.flags |= DAF_CHAINED;
        else if (!strcmp(s, "noise") || !strcmp(s, "no") || !strcmp(s, "bnr"))
          ATX.flags |= DAF_NOISE;
        else if (!strcmp(s, "whitelist") || !strcmp(s, "wh"))
          ATX.flags |= DAF_WHITELIST;
        else if (!strncmp(s, "tb=", 3)) { 
          ATX.training_buffer = atoi(strchr(s, '=')+1);
          if (ATX.training_buffer < 0 || ATX.training_buffer > 10) {
            report_error("training buffer level specified is invalid");
            exit(EXIT_FAILURE);
          }
        }
        else
        {
          report_error_printf("unknown feature '%s'", s);
          exit(EXIT_FAILURE);
        }
                                                                                
        s = strtok(NULL, ",");
      }
      free(d);
      continue;
    }

    /* Output message to stdout */
    if (!strcmp (argv[i], "--stdout"))
    {
      ATX.flags |= DAF_STDOUT;
      continue;
    }

    /* Append all other arguments as mailer args */
    if (i > 0 && !user_flag
#ifdef TRUSTED_USER_SECURITY
         && trusted
#endif
    )
    {
      if (argv[i][0] == 0)
        strlcat (ATX.mailer_args, "\"\"", sizeof (ATX.mailer_args));
      else
        strlcat (ATX.mailer_args, argv[i], sizeof (ATX.mailer_args));
      strlcat (ATX.mailer_args, " ", sizeof (ATX.mailer_args));
    }
  }
 

  if (ATX.classification != -1 && ATX.source == -1) 
  {
    report_error("A source must be specified for this classification");
    exit(EXIT_FAILURE);
  }

  if (ATX.source != -1 && ATX.classification == -1)
  {
    report_error("A source was specified, but no classification");
    exit(EXIT_FAILURE);
  }

  if (ATX.operating_mode == -1)
  {
    report_error("No operating mode was specified");
    exit(EXIT_FAILURE);
  }

  if (ATX.training_mode == -1)
  {
    report_error("No training mode was specified, you insensitive clod!");
    exit(EXIT_FAILURE);
  }

#ifdef QLOCAL
  snprintf (ATX.spam_args, sizeof (ATX.spam_args), "%s ", QLOCAL);
#endif

#ifndef PARSE_TO
  if (users->items == 0)
  {
    LOG (LOG_ERR, ERROR_USER_UNDEFINED);
    report_error (ERROR_USER_UNDEFINED);
    fprintf (stderr, "%s\n", SYNTAX);
    exit (EXIT_FAILURE);
  }
#endif

#ifdef TRUSTED_USER_SECURITY
  if (!trusted)
  {
    char filename[MAX_FILENAME_LENGTH];
    FILE *file;

    /* Use the arguments specifically assigned for nontrusted users */

    snprintf (filename, MAX_FILENAME_LENGTH,
              "%s/untrusted.mailer_args", DSPAM_HOME);
    file = fopen (filename, "r");
    if (file != NULL)
    {
      if (fgets (buff, sizeof (buff), file) != NULL)
      {
        ALLTRIM (buff);
        if (buff[0])
          strlcpy (ATX.mailer_args, buff, sizeof(ATX.mailer_args));
      }
      fclose (file);
    }
    else
    {
      report_error_printf(
           "unable to locate %s/untrusted.mailer_args: %s.",
           DSPAM_HOME, strerror (errno));
      exit(EXIT_FAILURE);
    }
  }
#endif

  /* Read in the message */
#ifdef PARSE_TO
  i = 0;
#endif
  while ((fgets (buff, sizeof (buff), stdin)) != NULL)
  {
    if (line > 1 || strncmp (buff, "From QUARANTINE", 15))
    {
#ifdef PARSE_TO
      /* Parse the To: address for a username */
      if (buff[0] == 0)
        i = 1;
      if (ATX.classification != -1 && !i && !strncasecmp(buff, "To: ", 4)) {
        char *y = NULL;
        if (ATX.classification == DSR_ISSPAM) {
          char *x = strstr(buff, "spam-");
          if (x != NULL) {
            y = strdup(x+5);
          }
        } else {
          char *x = strstr(buff, "fp-");
          if (x != NULL) {
            y = strdup(x+3);
          }
        }

        if (y) {
          char *z = strtok(y, "@");
          nt_add (users, z);
          free(y);
        }
      }
#endif
      if (buffer_cat (message, buff))
      {
        LOG (LOG_CRIT, ERROR_MEM_ALLOC);
        exit (EXIT_FAILURE);
      }
    }

    /* Use the original user id if we are reversing a false positive */
    if (!strncasecmp (buff, "X-DSPAM-User: ", 14) && 
        ATX.managed_group[0] != 0                 &&
        ATX.operating_mode == DSM_PROCESS    &&
        ATX.classification == DSR_ISINNOCENT &&
        ATX.source         == DSS_ERROR)
    {
      char user[MAX_USERNAME_LENGTH];
      strlcpy (user, buff + 14, sizeof (user));
      chomp (user);
      nt_destroy (users);
      users = nt_create (NT_CHAR);
      if (users == NULL)
      {
        report_error (ERROR_MEM_ALLOC);
        exit (EXIT_FAILURE);
      }
      nt_add (users, user);
    }

    line++;
  }

#ifdef PARSE_TO
  if (users->items == 0)
  {
    LOG (LOG_ERR, ERROR_USER_UNDEFINED);
    report_error (ERROR_USER_UNDEFINED);
    fprintf (stderr, "%s\n", SYNTAX);
    exit (EXIT_FAILURE);
  }
#endif


  if (!message->used)
  {
    LOG (LOG_INFO, "empty message (no data received at all)");
    exit (EXIT_SUCCESS);
  }

  if (dspam_init_driver ())
  {
    LOG (LOG_WARNING, "unable to initialize storage driver");
    exit (EXIT_FAILURE);
  }

  /* Process message for each user */
  node_nt = c_nt_first (users, &c_nt);
  while (node_nt != NULL)
  {
    PREF *PTX;
    struct stat s;
    char filename[MAX_FILENAME_LENGTH];
    int result;


#ifdef DEBUG
    DO_DEBUG = 0;
    strcpy (filename, _ds_userdir_path (node_nt->ptr, "debug"));
    if (!stat (filename, &s)) 
      DO_DEBUG = 1;
    else {
      snprintf(filename, sizeof(filename),"%s/.debug", DSPAM_HOME);
      if (!stat(filename, &s))
        DO_DEBUG = 1;
    }
 
    if (DO_DEBUG) {
      DO_DEBUG = 1;

      LOGDEBUG ("DSPAM Instance Startup");
      LOGDEBUG ("full arg list: %s", debug_args);

      LOGDEBUG ("pass-thru arguments: %s", ATX.mailer_args);
      LOGDEBUG ("processing user %s", (const char *) node_nt->ptr);

      /* Write message to DSPAM_HOME/dspam.messags */
      {
        FILE *f;
        char m[MAX_FILENAME_LENGTH];
        snprintf (m, sizeof (m), "%s/dspam.messages", DSPAM_HOME);
        f = fopen (m, "a");
        if (f != NULL)
        {
          fprintf (f, "%s\n", message->data);
          fclose (f);
        }
      }
    }
#endif

    LOGDEBUG("Loading preferences for user %s", (const char *) node_nt->ptr);
    PTX = load_preferences(node_nt->ptr);

#ifdef OPT_IN
    strcpy (filename, _ds_userdir_path (node_nt->ptr, "dspam"));
#else
    strcpy (filename, _ds_userdir_path (node_nt->ptr, "nodspam"));
#endif

#ifdef OPT_IN
    if (stat (filename, &s))
#else
    if (!stat (filename, &s))
#endif
    {
      if (ATX.flags & DAF_DELIVER_INNOCENT)
      {
        retcode =
          deliver_message (message->data,
                           (ATX.flags & DAF_STDOUT) ? NULL : ATX.mailer_args,
                            node_nt->ptr);
        if (retcode && exitcode == EXIT_SUCCESS)
          exitcode = retcode;
      }
    }

    /* Process/Classify Message */

    else
    {
      /* Drop a .corpuslock file to identify a corpus message is being 
         processed */

      if (ATX.source == DSS_CORPUS) {
        char lock[MAX_FILENAME_LENGTH];
        FILE *file;

        strcpy(lock, _ds_userdir_path (node_nt->ptr, "corpuslock"));
        file = fopen(lock, "w");
        if (file != NULL) {
          fprintf(file, "%ld", (long) getpid());
          fclose(file);
        }

        result = process_message (&ATX, PTX, message, node_nt->ptr);

        unlink(lock);
      }
      else 
      {
        result = process_message (&ATX, PTX, message, node_nt->ptr);
#ifdef BROKEN_RETURN_CODES
        if (result == DSR_ISSPAM)
          exitcode = 99;
#endif
      }

      /* Classify Only */

      if (ATX.operating_mode == DSM_CLASSIFY)
      {
        node_nt = c_nt_next (users, &c_nt);
        continue;
      }

      /* Classify and Process */

      /* Innocent */

      if (result != DSR_ISSPAM)
      {
        /* Processing Error */

        if (result != DSR_ISINNOCENT)
        {
          LOG (LOG_WARNING,
               "process_message returned error %d.  delivering message.",
               result);
        }

        /* Deliver */
        if (ATX.flags & DAF_DELIVER_INNOCENT) {
          LOGDEBUG ("delivering message");
            retcode = deliver_message
              (message->data,
               (ATX.flags & DAF_STDOUT) ? NULL : ATX.mailer_args,
               node_nt->ptr);
            if (retcode)
              exitcode = retcode;
        }
      }

      /* Spam */

      else
      {
        /* Do not Deliver Spam */

        if (! (ATX.flags & DAF_DELIVER_SPAM))
        {

          /* If a specific quarantine has been configured, use it */

          if (ATX.source != DSS_CORPUS) {
            if (ATX.spam_args[0] != 0 || 
                (PTX != NULL && (
                  PTX->spam_mode == DPS_TAG || PTX->spam_mode == DPS_DELIVER)))
            {
              if (ATX.spam_args[0] != 0) {
                retcode = deliver_message
                  (message->data,
                   (ATX.flags & DAF_STDOUT) ? NULL : ATX.spam_args, 
                   node_nt->ptr);
              } else {
                retcode = deliver_message
                  (message->data,
                   (ATX.flags & DAF_STDOUT) ? NULL : ATX.mailer_args,
                   node_nt->ptr);
              }
              if (retcode)
                exitcode = retcode;
            }
            else
            {
              /* Use standard quarantine procedure */
              if (ATX.source == DSS_INOCULATION || ATX.classification == -1) {
                if (ATX.managed_group[0] == 0)
                  quarantine_message (message->data, node_nt->ptr);
                else
                  quarantine_message (message->data, ATX.managed_group);
              }
            }
          }
        }

        /* Deliver Spam */

        else
        {
          retcode = deliver_message
            (message->data,
             (ATX.flags & DAF_STDOUT) ? NULL : ATX.mailer_args,
             node_nt->ptr);
          if (retcode)
            exitcode = retcode;
        }
      }
    }

    node_nt = c_nt_next (users, &c_nt);

    LOGDEBUG ("DSPAM Instance Shutdown.  Exit Code: %d", exitcode);

  }

  dspam_shutdown_driver ();
  exit (exitcode);
}

/*
   process_message: calls the dspam processor for each destination user. 
                    manages signatures.
           returns: DSR_ISINNOCENT - Message is innocent
                    DSR_ISSPAM     - Message is spam
                    (other)        - Error
*/

int
process_message (AGENT_CTX *ATX, 
                 PREF *PTX, 
                 buffer * message, 
                 const char *username)
{
  DSPAM_CTX *CTX;			/* dspam context */
  struct _ds_message *components;
  struct _ds_spam_signature SIG;        /* signature object */
  struct _ds_neural_decision DEC;	/* neural decision */
  char ctx_group[128];

  struct nt_node *node_nt;
  struct nt_c c_nt;

  FILE *file;
  char filename[MAX_FILENAME_LENGTH];
  char session[64] = { 0 };
  char *copyback;

  int have_signature = 0;
  int have_decision = 0;
  int f_all = 0;
  int f_mode = DSM_PROCESS;
  int result;
  int i = 0, was_spam = 0;
  u_int32_t operating_flags = 0;

  struct nt *inoc_users;        /* inoculate list */
  struct nt *classify_users;    /* classify list */
  ctx_group[0] = 0;
  DEC.length = 0;

  inoc_users = nt_create (NT_CHAR);
  if (inoc_users == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    return EUNKNOWN;
  }

  classify_users = nt_create (NT_CHAR);
  if (classify_users == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    return EUNKNOWN;
  }

  /* Set Group Membership */

  snprintf (filename, sizeof (filename), "%s/group", DSPAM_HOME);
  file = fopen (filename, "r");
  if (file != NULL)
  {
    char *group;
    char *user;
    char buffer[10240];

    while (fgets (buffer, sizeof (buffer), file) != NULL)
    {
      int do_inocgroups = 0;
      char *type, *list;
      chomp (buffer);
      list = strdup (buffer);
      group = strtok (buffer, ":");
      if (group != NULL)
      {
        type = strtok (NULL, ":");
        user = strtok (NULL, ",");

        if (!strcasecmp (type, "INOCULATION") &&
            ATX->classification == DSR_ISSPAM &&
            ATX->source != DSS_CORPUS)
        {
          do_inocgroups = 1;
        }

        while (user != NULL)
        {
          if (!strcmp (user, username) || user[0] == '*' ||
             (user[0] == '@' && !strcmp(user, strchr(username,'@'))))
          {

            /* If we're reporting a spam, report it as a spam to all other
             * users in the inoculation group */
            if (do_inocgroups)
            {
              char *l = list, *u;
              u = strsep (&l, ":");
              u = strsep (&l, ":");
              u = strsep (&l, ",");
              while (u != NULL)
              {
                if (strcmp (u, username))
                {
                  LOGDEBUG ("adding user %s to inoculation group %s", u,
                            group);
                 if (u[0] == '*') {
                    nt_add (inoc_users, u+1);
                  } else
                    nt_add (inoc_users, u);
                }
                u = strsep (&l, ",");
              }
            }
            else if (!strncasecmp (type, "SHARED", 6)) 
            {
              strlcpy (ctx_group, group, sizeof (ctx_group));
              LOGDEBUG ("assigning user %s to group %s", username, group);

              if (!strncasecmp (type + 6, ",MANAGED", 8))
                strlcpy (ATX->managed_group, 
                         ctx_group, 
                         sizeof(ATX->managed_group));

            }
            else if (!strcasecmp (type, "CLASSIFICATION") ||
                     !strcasecmp (type, "NEURAL")) 
            {
              char *l = list, *u;
              u = strsep (&l, ":");
              u = strsep (&l, ":");
              u = strsep (&l, ",");
              while (u != NULL)
              {
                if (strcmp (u, username))
                {
                  LOGDEBUG ("adding user %s to classification group %s", u,
                            group);
                  if (!strcasecmp (type, "NEURAL"))
                  operating_flags |= DAF_NEURAL;
             
                  if (u[0] == '*') {
                    operating_flags |= DAF_GLOBAL;
                    nt_add (classify_users, u+1);
                  } else
                    nt_add (classify_users, u);
                }
                u = strsep (&l, ",");
              }
            }
            else if (!strcasecmp (type, "MERGED") && strcmp(group, username))
            {
              char *l = list, *u;
              u = strsep (&l, ":");
              u = strsep (&l, ":");
              u = strsep (&l, ",");
              while (u != NULL)
              {
                if (!strcmp (u, username) || u[0] == '*')
                {
                    LOGDEBUG ("adding user to merged group %s", group);

                    ATX->flags |= DAF_MERGED;
                                                                                
                    strlcpy(ctx_group, group, sizeof(ctx_group));
                }
                u = strsep (&l, ",");
              }
            }
          }
          do_inocgroups = 0;
          user = strtok (NULL, ",");
        }
      }

      free (list);
    }
    fclose (file);
  }

  /* Crunch our agent context into a DSPAM context */

  f_mode = ATX->operating_mode;
  f_all  = DSF_SIGNATURE;

  if (ATX->flags & DAF_CHAINED)
    f_all |= DSF_CHAINED;

  /* If there is no preference, defer to commandline */
  if (PTX != NULL && PTX->feature_bnr != -1) {
    if (PTX->feature_bnr)
      f_all |= DSF_NOISE;
  } else {
    if (ATX->flags & DAF_NOISE)
     f_all |= DSF_NOISE;
  }

  if (PTX != NULL && PTX->feature_whitelist != -1) {
    if (PTX->feature_whitelist)
      f_all |= DSF_WHITELIST;
  } else {
    if (ATX->flags & DAF_WHITELIST)
      f_all |= DSF_WHITELIST;
  }

  if (ATX->flags & DAF_MERGED)
    f_all |= DSF_MERGED;

  CTX = dspam_init (username, ctx_group, f_mode, f_all);

  if (CTX == NULL)
  {
    LOG (LOG_WARNING, "unable to initialize dspam context");
    return EUNKNOWN;
  }

  if (PTX != NULL && PTX->training_buffer >= 0) 
    CTX->training_buffer = PTX->training_buffer;

  if (ATX->classification != -1) {
    CTX->classification  = ATX->classification;
    CTX->source	         = ATX->source;
  }

  if (PTX != NULL && PTX->training_mode != -1)
    CTX->training_mode = PTX->training_mode;
  else
    CTX->training_mode = ATX->training_mode;

  if (CTX->training_buffer != -1) 
    CTX->training_buffer = ATX->training_buffer;

  /* First Run Message */
  if (! CTX->totals.innocent_learned && ! CTX->totals.spam_learned) { 
    send_notice("firstrun.txt", ATX->mailer_args, username);
  }

  components = _ds_actualize_message (message->data);
  if (components == NULL)
  {
    LOG (LOG_WARNING, "_ds_actualize_message() failed.  unable to process.");
    return EUNKNOWN;
  }

  CTX->message = components;

  /* If this message is an inoculation, authenticate and process it */

  /* per Internet-Draft draft-yerazunis-inoculation-03.txt 
   * "A MIME Encoding for Message Inoculations"
   * 
   * UNDER DEVELOPMENT...
   * current caveats:
   * 
   * - only md5 authentication is supported
   * - only the message/inoculation MIME type is supported
   * - outgoing inoculation messages are not supported
   */

#ifdef EXPERIMENTAL

  {
    struct nt_node *node_nt, *node_hnt;
    struct nt_c c_nt;
    char filename[MAX_FILENAME_LENGTH];
    int auth_user = 0, auth_checksum = 0;
    char *inoc_message = NULL;
    FILE *file;

    node_nt = c_nt_first (CTX->message->components, &c_nt);
    if (node_nt != NULL)
    {
      struct _ds_message_block *block;

      block = (struct _ds_message_block *) node_nt->ptr;
      if (block->media_type == MT_MESSAGE
          && block->media_subtype == MST_INOCULATION)
      {
        struct _ds_header_field *header;
        char *sender = NULL, *checksum = NULL, *inoculation_type = NULL;
        int auth_type = IAT_UNKNOWN;
        long content_length = 0;

        LOGDEBUG ("validating inoculation");

        /* Read the inoculation headers */
        node_hnt = c_nt_first (block->headers, &c_nt);
        while (node_hnt != NULL)
        {
          header = (struct _ds_header_field *) node_hnt->ptr;
          if (header != NULL)
          {
            if (!strcasecmp (header->heading, "Content-Length"))
              content_length = strtol (header->data, NULL, 0);
            else if (!strcasecmp (header->heading, "Inoculation-Sender"))
              sender = header->data;
            else if (!strcasecmp
                     (header->heading, "Inoculation-Authentication"))
            {
              if (!strncasecmp (header->data, "md5", 3))
              {
                char *d = strdup (header->data);
                auth_type = IAT_MD5;

                if (strchr (header->data, '"'))
                {
                  checksum = strtok (d, "\"");
                  checksum = strdup (strtok (NULL, "\""));
                }
                else
                {
                  checksum = strtok (d, "=");
                  checksum = strdup (strtok (NULL, "="));
                }
                free (d);

              }
              else if (!strncasecmp (header->data, "signed", 6))
              {
                auth_type = IAT_SIGNED;
              }
              else if (!strncasecmp (header->data, "none", 4))
              {
                auth_type = IAT_NONE;
              }
            }
            else if (!strcasecmp (header->heading, "Inoculation-Type"))
              inoculation_type = header->data;
          }
          node_hnt = c_nt_next (block->headers, &c_nt);
        }

        /* Verify the inoculation */
        if (sender != NULL && checksum != NULL && inoculation_type != NULL)
        {
          char *s = NULL, *v = NULL;    /* s = sender, v = validation code */

          /* Verify the sender, get the validation code */
          strcpy (filename, _ds_userdir_path (username, "inoc"));
          file = fopen (filename, "r");
          if (file != NULL)
          {
            char buf[1024];
            while (fgets (buf, sizeof (buf), file) != NULL)
            {
              s = strdup (strtok (buf, ":"));
              v = strdup (strtok (NULL, ":"));
              if (!strcmp (s, sender))
              {
                LOGDEBUG ("sender verified");
                auth_user = 1;
              }
              else
              {
                free (s);
                free (v);
              }
            }
            fclose (file);
          }

          /* Verify the Checksum */
          if (auth_user)
          {
            if (auth_type == IAT_MD5)
            {
              unsigned char digest[17];
              char a_digest[33];
              char x[3];
              MD5_CTX MTX;

              if (content_length == 0)
                content_length = strlen (block->body->data) - 1;

              MD5Init (&MTX);
              MD5Update (&MTX, v, strlen (v));  /* Verification Code */
              MD5Update (&MTX, block->body->data, content_length);

              MD5Final (digest, &MTX);
              a_digest[0] = 0;
              for (i = 0; i < 16; i++)
              {
                snprintf (x, 3, "%02x", digest[i]);
                strcat (a_digest, x);
              }
              if (!strcmp (a_digest, checksum))
              {
                LOGDEBUG ("inoculation validated");
                auth_checksum = 1;
                inoc_message = strdup (block->body->data);
              }
              else
              {
                LOGDEBUG ("inoculation invalid. ignoring");
                return 0;
              }
            }
          }
        }
        free (checksum);
      }
    }

    /* Process the inoculation */
    if (auth_user && auth_checksum)
    {
      inoculate_user (username, NULL, inoc_message);
      free (inoc_message);
      return DSR_ISSPAM;
    }
  }

#endif

  /* END Message-Based Inoculation */

  /* Find, parse, and strip DSPAM signature */
  {
    char signature_key[128];
    struct nt_node *node_nt, *prev_node = NULL;
    struct nt_c c;
    struct _ds_message_block *block = NULL;
    char first_boundary[512];
    int is_signed = 0;

#ifdef SIGNATURE_IN_ATTACHMENTS
    int final_boundary = 0;
#else
#ifndef WEBMAIL
    char *signature_begin = NULL, *signature_end, *erase_begin;
    int signature_length;
#endif
    struct nt_node *node_header;
    struct nt_c c2;
#endif

    i = 0;
    first_boundary[0] = 0;

    /* Iterate through each message component in search of a signature
     * Decode components as necessary */

    node_nt = c_nt_first (CTX->message->components, &c);
    while (node_nt != NULL)
    {
      block = (struct _ds_message_block *) node_nt->ptr;

      if (block->media_type == MT_MULTIPART
          && block->media_subtype == MST_SIGNED)
        is_signed = 1;

#ifdef VERBOSE
      LOGDEBUG ("Scanning component %d for a DSPAM signature", i);
#endif

/* mode: signature-attachments */

#ifdef SIGNATURE_IN_ATTACHMENTS
      /* record the message's main boundary so we can make sure there isn't any junk in the last component */
      if (i == 0 && block->boundary != NULL)
        strlcpy (first_boundary, block->boundary, sizeof (first_boundary));

      /* strip out any junk block that came after the end of the message */

      if (final_boundary)
      {
        struct nt_node *old;
        _ds_destroy_block (block);
        if (prev_node != NULL)
          prev_node->next = node_nt->next;
        else
          CTX->message->components->first = node_nt->next;

        CTX->message->components->insert = NULL;

        old = node_nt;
        node_nt = node_nt->next;
        free (old);
        CTX->message->components->items--;
        continue;
      }

      /* if this block terminates the message, set our final boundary */
      if (block->terminating_boundary != NULL && first_boundary[0] != 0)
      {
        char boundary[512];
        snprintf (boundary, sizeof (boundary), "%s--", first_boundary);
        if (!strcmp (block->terminating_boundary, boundary))
          final_boundary = 1;
      }

      /* Signature found in attachment; reassemble into _ds_spam_signature */
      if (block->media_type == MT_APPLICATION
          && block->media_subtype == MST_DSPAM_SIGNATURE)
      {
        char *token, *attachment, *m_attachment;
        long num_tokens, i = 0;
        unsigned long long value;

        attachment = strdup (block->body->data);
        m_attachment = attachment;
        token = strsep (&attachment, "\n");
        num_tokens = strtol (token, NULL, 0);
        LOGDEBUG ("found dspam.dat; %ld tokens to reverse", num_tokens);

        SIG.length = num_tokens * sizeof (unsigned long long);
        SIG.data = malloc (num_tokens * sizeof (unsigned long long));

        if (SIG.data == NULL)
        {
          LOG (LOG_CRIT, ERROR_MEM_ALLOC);
          return EUNKNOWN;
        }

        token = strsep (&attachment, "\n");
        while (token != NULL && i < num_tokens)
        {
          value = strtoull (token, NULL, 16);
          if (value > 0)
          {
            memcpy (SIG.data +
                    (i *
                     sizeof (unsigned long
                             long)), &value, sizeof (unsigned long long));
          }
          token = strsep (&attachment, "\n");
          i++;
        }

        CTX->signature = &SIG;
        have_signature = 2;
        free (m_attachment);

        /* Strip out dspam data attachment */
        if (prev_node != NULL)
        {
          struct _ds_message_block *prev_block =
            (struct _ds_message_block *) prev_node->ptr;

          /* use the current block's terminating boundary, since we're removing it */
          free (prev_block->terminating_boundary);
          prev_block->terminating_boundary = block->terminating_boundary;
          block->terminating_boundary = NULL;

          prev_node->next = node_nt->next;
          node_nt = prev_node;
          CTX->message->components->insert = NULL;
          CTX->message->components->items--;
        }
        else
        {
          CTX->message->components->first = node_nt->next;
          CTX->message->components->items = 1;
          CTX->message->components->insert = node_nt->next;
          node_nt = NULL;
        }

        _ds_destroy_block (block);

      }
#else
      if (block->media_type == MT_TEXT
          || block->media_type == MT_MESSAGE 
          || block->media_type == MT_UNKNOWN ||
                  (i == 0 && (block->media_type == MT_TEXT ||
                              block->media_type == MT_MULTIPART ||
                              block->media_type == MT_MESSAGE) ))
      {
        char *body;

        /* Verbose output of each message component */
#ifdef VERBOSE
        if (DO_DEBUG) {
          if (block->boundary != NULL)
          {
            LOGDEBUG ("  : Boundary     : %s", block->boundary);
          }
          if (block->terminating_boundary != NULL)
            LOGDEBUG ("  : Term Boundary: %s", block->terminating_boundary);
          LOGDEBUG ("  : Encoding     : %d", block->encoding);
          LOGDEBUG ("  : Media Type   : %d", block->media_type);
          LOGDEBUG ("  : Media Subtype: %d", block->media_subtype);
          LOGDEBUG ("  : Headers:");
          node_header = c_nt_first (block->headers, &c2);
          while (node_header != NULL)
          {
            struct _ds_header_field *header =
              (struct _ds_header_field *) node_header->ptr;
            LOGDEBUG ("    %-32s  %s", header->heading, header->data);
            node_header = c_nt_next (block->headers, &c2);
          }
        }
#endif

        body = block->body->data;
        if (block->encoding == EN_BASE64
            || block->encoding == EN_QUOTED_PRINTABLE)
        {
          struct _ds_header_field *field;
          int is_attachment = 0;
          struct nt_node *node_hnt;
          struct nt_c c_hnt;

          node_hnt = c_nt_first (block->headers, &c_hnt);
          while (node_hnt != NULL)
          {
            field = (struct _ds_header_field *) node_hnt->ptr;
            if (field != NULL
                && field->heading != NULL && field->data != NULL)
              if (!strncasecmp (field->heading, "Content-Disposition", 19))
                if (!strncasecmp (field->data, "attachment", 10))
                  is_attachment = 1;
            node_hnt = c_nt_next (block->headers, &c_hnt);
          }

          if (!is_attachment)
          {
#ifdef VERBOSE
            LOGDEBUG ("decoding message block from encoding type %d",
                      block->encoding);
#endif

            body = _ds_decode_block (block);

            if (is_signed)
            {
              LOGDEBUG
                ("message is signed.  retaining original text for reassembly");
              block->original_signed_body = block->body;
            }
            else
            {
              block->encoding = EN_7BIT;

              node_header = c_nt_first (block->headers, &c2);
              while (node_header != NULL)
              {
                struct _ds_header_field *header =
                  (struct _ds_header_field *) node_header->ptr;
                if (!strcasecmp
                    (header->heading, "Content-Transfer-Encoding"))
                {
                  free (header->data);
                  header->data = strdup ("7bit");
                }
                node_header = c_nt_next (block->headers, &c2);
              }

              buffer_destroy (block->body);
            }
            block->body = buffer_create (body);
            free (body);

            body = block->body->data;
          }
        }

#ifdef SIGNATURE_IN_HEADERS
        if (block->headers != NULL && !have_signature)
        {
          struct nt_node *node_header;
          struct _ds_header_field *head;

          node_header = block->headers->first;
          while(node_header != NULL) {
            head = (struct _ds_header_field *) node_header->ptr;
            if (head->heading && !strcmp(head->heading, "X-DSPAM-Signature")) {
              body = head->data;
              break;
            } 
            node_header = node_header->next;
          }
        }
#endif
#ifndef WEBMAIL
        /* Look for signature */
        if (body != NULL)
        {
          signature_begin = strstr (body, SIGNATURE_BEGIN);
          if (signature_begin == NULL)
            signature_begin = strstr (body, LOOSE_SIGNATURE_BEGIN);

          if (signature_begin != NULL)
          {
            erase_begin = signature_begin;
            signature_begin =
              strstr
              (signature_begin,
               SIGNATURE_DELIMITER) + strlen (SIGNATURE_DELIMITER);
            signature_end = signature_begin;

            /* Find the signature's end character */
            while (signature_end != NULL
                   && signature_end[0] !=
                   0
                   &&
                   (isalnum
                    ((int) signature_end[0]) || signature_end[0] == 32))
              signature_end++;

            if (signature_end != NULL)
            {
              signature_length = signature_end - signature_begin;

              if (signature_length < 128)
              {
                memcpy (signature_key, signature_begin, signature_length);
                signature_key[signature_length] = 0;
#ifndef SIGNATURE_IN_HEADERS
                if (!is_signed && ATX->classification == -1) {
                  strcpy (erase_begin, signature_end + 1);
                  block->body->used = strlen(body);
                }
#endif
                have_signature = 1;
                LOGDEBUG ("found signature '%s'", signature_key);
              }
            }
          }
        }
#endif /* !WEBMAIL */
      }
#endif
      prev_node = node_nt;
      node_nt = c_nt_next (components->components, &c);
      i++;
    }
    if (have_signature == 1)
    {
      have_decision = 1;
      if (_ds_get_signature (CTX, &SIG, signature_key))
      {
        LOGDEBUG ("signature retrieval for '%s' failed", signature_key);
        have_signature = 0;
      }
      else
      {
        CTX->signature = &SIG;
//        _ds_delete_signature (CTX, signature_key);
      }

#ifdef NEURAL
      if (_ds_get_decision (CTX, &DEC, signature_key))
        have_decision = 0;
//      else
//        _ds_delete_decision (CTX, signature_key);
#endif
    }
  }

  /* Set neural node reliability based on which nodes classified the message
     correctly and incorrectly */

#ifdef NEURAL
  if (have_decision                   &&
      CTX->classification != DSR_NONE && 
      CTX->source == DSS_ERROR)
  {
    struct _ds_neural_record r;
    char d;
    void *ptr;

    for(ptr = DEC.data;ptr<DEC.data+DEC.length;ptr+=sizeof(uid_t)+1) {
       memcpy(&r.uid, ptr, sizeof(uid_t));
       memcpy(&d, ptr+sizeof(uid_t), 1);
       if ((d == 'S' && CTX->classification == DSR_ISINNOCENT) ||
           (d == 'I' && CTX->classification == DSR_ISSPAM)) {
         if (!_ds_get_node(CTX, NULL, &r)) {
           r.total_incorrect++;
           r.total_correct--;
           _ds_set_node(CTX, NULL, &r);
         }
       }
    }
    free(DEC.data);
  }
#endif

  /* Unset DSF_SIGNATURE flag if we don't have a signature and we're either
     classifying the message or training */

  if (!have_signature && 
      (CTX->operating_mode == DSM_CLASSIFY || CTX->classification != DSR_NONE))
  {

    CTX->flags = CTX->flags ^ DSF_SIGNATURE;
    CTX->signature = NULL;
  }

  /* If we have the signature, and are either classifying or training... */

  if (have_signature &&
      (CTX->operating_mode == DSM_CLASSIFY || CTX->classification != DSR_NONE))
  {
#ifdef TEST_COND_TRAINING
    int do_train = 1, iter = 0, ck_result = 0, t_mode = CTX->source;

    if (f_mode == DSM_CLASSIFY)
      do_train = 0;

    /* train until test conditions are met, 5 iterations max */
    while (do_train && iter < 5)
    {
      DSPAM_CTX *CLX;
      int match = (CTX->classification == DSR_ISSPAM) ? DSR_ISSPAM : DSR_ISINNOCENT;

      iter++;
#endif

      result = dspam_process (CTX, NULL);

#ifdef TEST_COND_TRAINING

      /* Only subtract innocent values once */
      CTX->source = DSS_CORPUS;

      LOGDEBUG ("reclassifying iteration %d result: %d", iter, result);

      if (t_mode == DSS_CORPUS)
        do_train = 0;

      /* only attempt test-conditional training on a mature corpus */
      if (CTX->totals.innocent_learned + CTX->totals.innocent_classified 
          < 1000 && 
            CTX->classification == DSR_ISSPAM)
        do_train = 0;
      else
      {
        int f_all =  DSF_SIGNATURE;

        /* CLX = Classify Context */
#ifdef CHAINED_TOKENS
        f_all = f_all ^ DSF_CHAINED;
#endif

#ifdef BNR
        f_all = f_all ^ DSF_NOISE;
#endif

        CLX = dspam_init (CTX->username, CTX->group, DSM_CLASSIFY, f_all);
        if (!CLX)
        {
          do_train = 0;
          break;
        }
        CLX->signature = &SIG;
        ck_result = dspam_process (CLX, NULL);
        if (ck_result || CLX->result == match)
          do_train = 0;
        CLX->signature = NULL;
        dspam_destroy (CLX);
      }
    }

    CTX->source = DSS_ERROR;

#endif
  }
  else
  {
    CTX->signature = NULL;
#ifndef WEBMAIL
    if (CTX->classification != DSR_NONE && CTX->source == DSS_ERROR) {
      dspam_destroy (CTX);
      LOGDEBUG("unable to find signature; bailing.");
      return EFAILURE;
    }
#endif
//      CTX->flags |= DSF_IGNOREHEADER;
    result = dspam_process (CTX, message->data);
  }

  result = CTX->result;

  /* First Spam Message */
  if (result == DSR_ISSPAM &&
        CTX->totals.spam_learned - CTX->totals.spam_misclassified - 
        CTX->totals.spam_corpusfed == 1) 
  {
    send_notice("firstspam.txt", ATX->mailer_args, username);
  }

  /* Quarantine Size Check */
  if (1) {
    struct stat s;
    char qfile[MAX_FILENAME_LENGTH];

    strcpy (qfile, _ds_userdir_path (username, "mbox"));
    if (!stat(qfile, &s) && s.st_size > 1024*1024*2) {
      strcpy (qfile, _ds_userdir_path (username, "mboxwarn"));
      if (stat(qfile, &s)) {
        FILE *f;

        f = fopen(qfile, "w");
        if (f != NULL) {
          fprintf(f, "%ld", (long) time(NULL));
          fclose(f);

          send_notice("quarantinefull.txt", ATX->mailer_args, username);
        }
      }
    }    
  }

  /* Defer to global group */
  if (operating_flags & DAF_GLOBAL && 
      ( (CTX->totals.innocent_learned + CTX->totals.innocent_corpusfed < 1000 || 
         CTX->totals.spam_learned + CTX->totals.spam_corpusfed < 250)         ||
      ( CTX->training_mode == DST_NOTRAIN))
     )
  {
    if (result == DSR_ISSPAM) { 
      was_spam = 1;
      CTX->result = DSR_ISINNOCENT;
      result = DSR_ISINNOCENT;
    }
    CTX->confidence = 0.60;
  }

  if (result != DSR_ISSPAM               && 
      CTX->operating_mode == DSM_PROCESS &&
      CTX->classification == DSR_NONE && 
      CTX->confidence < 0.65) 
  {

#ifdef NEURAL
    /* Consult neural network */
    if (ATX->flags & DAF_NEURAL) {
      struct _ds_neural_record r;
      struct nt_node *node_int;
      struct nt_c c_i;
      struct tbt *t;
      struct tbt_node *node_tbt;
      int total_nodes = classify_users->items;
      float bay_top = 0.0;
      float bay_bot = 0.0;
      float bay_result;
      int bay_used = 0;
      int res, i = 0;

      DEC.length = (total_nodes * (sizeof(uid_t)+1));
      DEC.data = calloc(1, DEC.length);
      
      if (DEC.data == NULL) {
        LOG(LOG_CRIT, ERROR_MEM_ALLOC);
        return EUNKNOWN;
      }

      t = tbt_create();
      t->type = 1; /* value based */
      
      node_int = c_nt_first (classify_users, &c_i);
      while (node_int != NULL) 
      {
        res = _ds_get_node(CTX, node_int->ptr, &r);
        if (!res) {
          memcpy(DEC.data+(i*(sizeof(uid_t)+1)), &r.uid, sizeof(uid_t));
          memset(DEC.data+(i*(sizeof(uid_t)+1))+sizeof(uid_t), 'E', 1);
       
          LOGDEBUG ("querying node %s", (const char *) node_int->ptr);
          res = user_classify ((const char *) node_int->ptr,
                                  CTX->signature, NULL);

          if ((res == DSR_ISSPAM || res == DSR_ISINNOCENT) && r.total_incorrect+r.total_correct>4) {
            tbt_add (t, 
                     (double) r.total_correct / (r.total_correct+r.total_incorrect),
                     r.uid, res, 0);
          }
          r.total_correct++;
          _ds_set_node(CTX, NULL, &r);

          if (res == DSR_ISSPAM)
            memset(DEC.data+(i*(sizeof(uid_t)+1))+sizeof(uid_t), 'S', 1);
          else if (res == DSR_ISINNOCENT)
            memset(DEC.data+(i*(sizeof(uid_t)+1))+sizeof(uid_t), 'I', 1);

          i++;
        }
        node_int = c_nt_next (classify_users, &c_i);
      }

      total_nodes /= 5;
      if (total_nodes<2)
        total_nodes = 2;
      node_tbt = tbt_first (t);

      /* include the top n reliable sources */
      while(node_tbt != NULL && total_nodes>0) {
        float probability = (node_tbt->frequency == DSR_ISINNOCENT) ? 1-node_tbt->probability : node_tbt->probability;

        LOGDEBUG("including node %llu [%2.6f]", node_tbt->token, probability);
        if (bay_used == 0)
        {
          bay_top = probability;
          bay_bot = 1 - probability; 
        }
        else
        {
          bay_top *= probability;
          bay_bot *= (1 - probability);
        }
      
        bay_used++;
        total_nodes--;
        node_tbt = tbt_next(node_tbt);
      }

      if (bay_used) { 
        bay_result = (bay_top) / (bay_top + bay_bot);
        if (bay_result > 0.80) { 
          result = DSR_ISSPAM;
        }

        LOGDEBUG("Neural Network Result: %2.6f", bay_result); 
      }

    /* Consult classification network */
    } else {
#endif
      struct nt_node *node_int;
      struct nt_c c_i;

      node_int = c_nt_first (classify_users, &c_i);
      while (node_int != NULL && result != DSR_ISSPAM)
      {
        LOGDEBUG ("checking result for user %s", (const char *) node_int->ptr);
        result = user_classify ((const char *) node_int->ptr,
                                CTX->signature, NULL);
        if (result == DSR_ISSPAM)
        {
          LOGDEBUG ("CLASSIFY CATCH: %s", (const char *) node_int->ptr);
          CTX->result = result;
        }
  
        node_int = c_nt_next (classify_users, &c_i);
      }
#ifdef NEURAL
    }
#endif

    /* Re-add as spam */
    if (result == DSR_ISSPAM && !was_spam)
    {
      DSPAM_CTX *CTC = malloc(sizeof(DSPAM_CTX));

      if (CTC == NULL) {
        report_error(ERROR_MEM_ALLOC);
        return -1;
      }

      memcpy(CTC, CTX, sizeof(DSPAM_CTX));

      CTC->operating_mode = DSM_PROCESS;
      CTC->classification = DSR_ISSPAM;
      CTC->source         = DSS_ERROR;
      CTC->flags         |= DSF_SIGNATURE;
      dspam_process (CTC, NULL);
      memcpy(&CTX->totals, &CTC->totals, sizeof(struct _ds_spam_totals));
      free(CTC);
      CTX->totals.spam_misclassified--;
      CTX->result = result;
    }

    /* If the global user thinks it's innocent, and the user thought it was
       spam, retrain the user as a false positive */
    if (result == DSR_ISINNOCENT && was_spam) {
      DSPAM_CTX *CTC = malloc(sizeof(DSPAM_CTX));
      if (CTC == NULL) {
        report_error(ERROR_MEM_ALLOC);
        return -1;
      }
                                                                                
      memcpy(CTC, CTX, sizeof(DSPAM_CTX));
                                                                                
      CTC->operating_mode = DSM_PROCESS;
      CTC->classification = DSR_ISINNOCENT;
      CTC->source         = DSS_ERROR;
      CTC->flags         |= DSF_SIGNATURE;
      dspam_process (CTC, NULL);
      memcpy(&CTX->totals, &CTC->totals, sizeof(struct _ds_spam_totals));
      free(CTC);
      CTX->totals.innocent_misclassified--;
      CTX->result = result;
    }
  }

  /* Inoculate other users: Signature */
  if (have_signature                   && 
     CTX->classification == DSR_ISSPAM && 
     CTX->source != DSS_CORPUS         && 
     inoc_users->items > 0)
  {
    struct nt_node *node_int;
    struct nt_c c_i;


    node_int = c_nt_first (inoc_users, &c_i);
    while (node_int != NULL)
    {
      inoculate_user ((const char *) node_int->ptr, &SIG, NULL);
      node_int = c_nt_next (inoc_users, &c_i);
    }
  }

  /* Inoculate other users: Message */
  if (!have_signature                   && 
      CTX->classification == DSR_ISSPAM &&
      CTX->source != DSS_CORPUS         &&
      inoc_users->items > 0)
  {
    struct nt_node *node_int;
    struct nt_c c_i;
    node_int = c_nt_first (inoc_users, &c_i);
    while (node_int != NULL)
    {
      inoculate_user ((const char *) node_int->ptr, NULL, message->data);
      node_int = c_nt_next (inoc_users, &c_i);
    }
    inoculate_user (username, NULL, message->data);
    return DSR_ISSPAM;
  }

  /* Save the DSPAM signature */
  if (CTX->operating_mode == DSM_PROCESS &&
      CTX->classification == DSR_NONE    &&
      CTX->signature != NULL)
  {
    int valid = 0;

    while (valid == 0)
    {
      _ds_create_signature_id (CTX, session, sizeof (session));
#ifndef SIGNATURE_IN_ATTACHMENTS
      if (_ds_verify_signature (CTX, session))
#endif
        valid = 1;
    }

    LOGDEBUG ("saved signature as %s", session);

#ifdef SIGNATURE_IN_ATTACHMENTS

    /* Embed the signature as an attachment; convert email if necessary */

    node_nt = c_nt_first (CTX->message->components, &c_nt);
    if (node_nt != NULL)
    {
      struct _ds_message_block *block = node_nt->ptr;
      if (block != NULL && block->headers != NULL)
      {
        long num_tokens, i;
        unsigned long long token;
        char boundary[128], t[128];
        buffer *s;
        int boundary_ok = 0;

        boundary[0] = 0;
        s = buffer_create (NULL);

        if (s == NULL)
        {
          LOG (LOG_CRIT, ERROR_MEM_ALLOC);
          return EUNKNOWN;
        }

        num_tokens = CTX->signature->length / sizeof (unsigned long long);
        snprintf (t, sizeof (t), "%ld\n", num_tokens);
        buffer_cat (s, t);

        for (i = 0; i < num_tokens; i++)
        {
          memcpy (&token,
                  CTX->signature->data +
                  (i *
                   sizeof (unsigned long long)), sizeof (unsigned long long));
          snprintf (t, sizeof (t), "%016llx", token);

          buffer_cat (s, t);
          buffer_cat (s, "\n");
        }

        buffer_cat (s, "\n");

        /* Message is not multipart; convert into multipart:
         * 
         * - Rewrite message header; add multipart boundary
         * - Move body to new block with original headers
         * 
         */

        if (block->media_type != MT_MULTIPART)
        {
          struct _ds_header_field *header,
            *original_header = NULL, *original_encoding = NULL;
          struct _ds_message_block *newblock;
          struct nt_node *nt_header, *prev = NULL;
          struct nt_c c_header;

          boundary_ok = 1;
          snprintf (boundary, sizeof (boundary), "=ds_%s=", session);

          /* Rewrite message header */

          nt_header = c_nt_first (block->headers, &c_header);
          while (nt_header != NULL)
          {
            char typedata[512];
            header = (struct _ds_header_field *) nt_header->ptr;
            if (!strcasecmp (header->heading, "Content-Type"))
            {
              original_header = header;

              snprintf (typedata, sizeof (typedata),
                   "Content-Type: multipart/mixed; boundary=\"%s\"", boundary);

              nt_header->ptr = (void *) _ds_create_header_field(typedata);
              if (nt_header->ptr == NULL)
              {
                LOG (LOG_CRIT, ERROR_MEM_ALLOC);
                return EUNKNOWN;
              }

              /* Replace original header */

              header = (struct _ds_header_field *) nt_header->ptr;
              block->terminating_boundary = strdup (boundary);
            }

            /* Remove Encoding */
            if (!strcasecmp (header->heading, "Content-Transfer-Encoding"))
            {
              if (prev == NULL)
                block->headers->first = nt_header->next;
              else {
                prev->next = nt_header->next;
                if (nt_header->next == NULL)
                  block->headers->insert = prev;
              }

              original_encoding = header;
              block->headers->items--;
            }

            prev = nt_header;
            nt_header = c_nt_next (block->headers, &c_header);
          }

          /* Add the header if there isn't one to replace */
          if (original_header == NULL)
          {
            char typedata[512];

            snprintf (typedata, sizeof (typedata),
                 "Content-Type: multipart/mixed; boundary=\"%s\"", boundary);

            header = _ds_create_header_field(typedata);
            if (header == NULL)
            {
              LOG (LOG_CRIT, ERROR_MEM_ALLOC);
              return EUNKNOWN;
            }
            nt_add (block->headers, (void *) header);
            block->terminating_boundary = strdup (boundary);
          }

          /* Create new message block */
          newblock =
            (struct _ds_message_block *)
            malloc (sizeof (struct _ds_message_block));
          if (newblock == NULL)
          {
            LOG (LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }

          newblock->headers = nt_create (NT_PTR);
          if (newblock->headers == NULL)
          {
            free (newblock);
            LOG (LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }

          /* Move block type info to new block */

          newblock->boundary = NULL;
          newblock->terminating_boundary = strdup (boundary);
          newblock->encoding = block->encoding;
          newblock->original_encoding = block->original_encoding;
          newblock->media_type = block->media_type;
          newblock->media_subtype = block->media_subtype;
          block->encoding = EN_UNKNOWN;
          block->original_encoding = EN_UNKNOWN;
          block->media_type = MT_MULTIPART;
          block->media_subtype = MST_MIXED;
          newblock->body = block->body;
          newblock->original_signed_body = NULL;
          block->body = buffer_create (NULL);

          if (original_header != NULL)
            nt_add (newblock->headers, original_header);
          if (original_encoding != NULL)
            nt_add (newblock->headers, original_encoding);

          /* Add the new block */

          nt_add (CTX->message->components, (void *) newblock);
        }

        /* Message is multipart/alternative; convert into multipart/mixed
         * 
         * ALTERNATIVE 
         * 
         * - Move multipart/alternative heading to new multipart block in message
         * - Rewrite message header as multipart/mixed
         * - Create new block with message body
         * 
         */

        else if (block->media_subtype == MST_ALTERNATIVE)
        {
          struct nt_node *new_node = malloc (sizeof (struct nt_node));
          struct _ds_header_field *header,
            *original_header = NULL, *original_encoding = NULL;
          struct _ds_message_block *newblock;
          struct nt_node *nt_header, *prev = NULL;
          struct nt_c c_header;
          char old_boundary[sizeof (boundary)];

          boundary_ok = 1;
          snprintf (boundary, sizeof (boundary), "=ds_%s=", session);
          strlcpy (old_boundary,
                   block->terminating_boundary, sizeof (old_boundary));

          /* Rewrite message header */

          nt_header = c_nt_first (block->headers, &c_header);
          while (nt_header != NULL)
          {
            char typedata[512];
            header = (struct _ds_header_field *) nt_header->ptr;
            if (!strcasecmp (header->heading, "Content-Type"))
            {
              original_header = header;
              snprintf (typedata, sizeof (typedata),
                  "Content-Type: multipart/mixed; boundary=\"%s\"", boundary);

              nt_header->ptr = _ds_create_header_field(typedata);
              if (nt_header->ptr == NULL)
              {
                LOG (LOG_CRIT, ERROR_MEM_ALLOC);
                return EUNKNOWN;
              }

              /* Replace original header */

              header = (struct _ds_header_field *) nt_header->ptr;
              block->terminating_boundary = strdup (boundary);
            }

            /* Remove Encoding */
            if (!strcasecmp (header->heading, "Content-Transfer-Encoding"))
            {
              if (prev == NULL)
                block->headers->first = nt_header->next;
              else
                prev->next = nt_header->next;

              original_encoding = header;
              block->headers->items--;
            }

            prev = nt_header;
            nt_header = c_nt_next (block->headers, &c_header);
          }

          /* Create new message block */
          newblock =
            (struct _ds_message_block *)
            malloc (sizeof (struct _ds_message_block));
          if (newblock == NULL)
          {
            LOG (LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }

          newblock->headers = nt_create (NT_PTR);
          if (newblock->headers == NULL)
          {
            free (newblock);
            LOG (LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }

          if (new_node == NULL)
          {
            LOG (LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }

          /* Move block type info to new block */
          newblock->boundary = NULL;
          newblock->terminating_boundary = strdup (old_boundary);
          newblock->encoding = block->encoding;
          newblock->original_encoding = block->original_encoding;
          newblock->media_type = block->media_type;
          newblock->media_subtype = block->media_subtype;
          block->encoding = EN_UNKNOWN;
          block->original_encoding = EN_UNKNOWN;
          block->media_type = MT_MULTIPART;
          block->media_subtype = MST_ALTERNATIVE;
          newblock->body = block->body;
          newblock->original_signed_body = NULL;
          block->body = buffer_create (NULL);

          if (original_header != NULL)
            nt_add (newblock->headers, original_header);
          if (original_encoding != NULL)
            nt_add (newblock->headers, original_encoding);

          /* Insert the new block before the alternative blocks */
          new_node->ptr = (void *) newblock;
          new_node->next = node_nt->next;
          node_nt->next = new_node;
          CTX->message->components->insert = NULL;

          /* Add a new terminating block */
          /* Create new message block */
          newblock =
            (struct _ds_message_block *)
            malloc (sizeof (struct _ds_message_block));
          if (newblock == NULL)
          {
            LOG (LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }

          newblock->headers = nt_create (NT_PTR);
          if (newblock->headers == NULL)
          {
            free (newblock);
            LOG (LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }

          newblock->boundary = NULL;
          newblock->terminating_boundary = strdup (boundary);
          newblock->encoding = EN_UNKNOWN;
          newblock->original_encoding = EN_UNKNOWN;
          newblock->media_type = EN_UNKNOWN;
          newblock->media_subtype = EN_UNKNOWN;
          newblock->body = buffer_create (NULL);
          newblock->original_signed_body = NULL;
          nt_add (CTX->message->components, (void *) newblock);

        }

        /* 
         * Message is now in an acceptable multipart format;
         * insert signature attachment 
         */

        /* Strip off the -- on the last terminating boundary */
        if (!boundary_ok)
        {
          char term[sizeof (boundary)];
          strlcpy (boundary, block->boundary, sizeof (boundary));
          snprintf (term, sizeof (term), "%s--", boundary);

          while (node_nt != NULL && term[0] != 0)
          {
            block = (struct _ds_message_block *) node_nt->ptr;
            if (block->
                terminating_boundary !=
                NULL && !strcmp (block->terminating_boundary, term))
            {
              free (block->terminating_boundary);
              block->terminating_boundary = strdup (boundary);
              term[0] = 0;
            }
            node_nt = c_nt_next (CTX->message->components, &c_nt);
          }
          node_nt = c_nt_first (CTX->message->components, &c_nt);
          block = (struct _ds_message_block *) node_nt->ptr;
        }

        if (boundary[0] != 0)
        {
          struct _ds_message_block *newblock;
          struct _ds_header_field *field;
          char term[512], typedata[512];

          /* Create new message block */
          newblock =
            (struct _ds_message_block *)
            malloc (sizeof (struct _ds_message_block));
          if (newblock == NULL)
          {
            LOG (LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }

          newblock->headers = nt_create (NT_PTR);
          if (newblock->headers == NULL)
          {
            free (newblock);
            LOG (LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }

          /* Create new block information */
          snprintf (term, sizeof (term), "%s--", boundary);
          newblock->boundary = NULL;
          newblock->terminating_boundary = strdup (term);
          newblock->encoding = EN_BASE64;
          newblock->original_encoding = EN_BASE64;
          newblock->media_type = MT_APPLICATION;
          newblock->media_subtype = MST_DSPAM_SIGNATURE;
          newblock->body = s;
          newblock->original_signed_body = NULL;

          snprintf(typedata, sizeof(typedata), 
              "Content-Type: application/dspam-signature; name=\"dspam.dat\"");

          field = _ds_create_header_field(typedata);
          if (field == NULL)
          {
            _ds_destroy_block (newblock);
            LOG (LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }
          nt_add (newblock->headers, field);

          field = (struct _ds_header_field *)
            malloc (sizeof (struct _ds_header_field));
          if (field == NULL)
          {
            _ds_destroy_block (newblock);
            LOG (LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }
          field->heading = strdup ("Content-Transfer-Encoding");
          field->data = strdup ("7bit");
          field->original_data = NULL;
          nt_add (newblock->headers, field);

          /* Add the new block */

          nt_add (CTX->message->components, (void *) newblock);

        }
      }
    }

#else
    if (CTX->classification == DSR_NONE && CTX->training_mode != DST_NOTRAIN) {
#ifndef WEBMAIL
      _ds_set_signature (CTX, CTX->signature, session);
#endif
#ifdef NEURAL
       if (DEC.length != 0)
        _ds_set_decision(CTX, &DEC, session);
#endif
    }
#endif
  }

  write_web_stats ((CTX->group == NULL || CTX->flags & DSF_MERGED) ? username : CTX->group,
                   &CTX->totals);

  LOGDEBUG ("libdspam returned probability of %f, message result: %s",
            CTX->probability, (result != DSR_ISSPAM) ? "NOT SPAM" : "SPAM");

#if defined(USER_LOGGING) || defined(SYSTEM_LOGGING)
  {
    char filename[MAX_FILENAME_LENGTH];
    char *subject = NULL, *from = NULL;
    FILE *file;
    char class;
    char x[1024];
    int y;

    strcpy (filename, _ds_userdir_path (username, "log"));

    node_nt = c_nt_first (CTX->message->components, &c_nt);
    if (node_nt != NULL)
    {
      struct _ds_message_block *block;
                                                                                
      block = node_nt->ptr;
      if (block->headers != NULL)
      {
        struct _ds_header_field *head;
        struct nt_node *node_header;

        node_header = block->headers->first;
        while(node_header != NULL) {
          head = (struct _ds_header_field *) node_header->ptr;
          if (head != NULL && !strcasecmp(head->heading, "Subject")) 
            subject = head->data;
          else if (head != NULL && !strcasecmp(head->heading, "From"))
            from = head->data;

          node_header = node_header->next;
        }
      }
    }

    if (result == DSR_ISSPAM)
      class = 'S';
    else
      class = 'I';

    if (CTX->probability == -2)
      class = 'W';

    if (CTX->source == DSS_ERROR) { 
      if (CTX->classification == DSR_ISSPAM)
        class = 'M';
      else if (CTX->classification == DSR_ISINNOCENT)
        class = 'F';
    }

    if (CTX->source == DSS_INOCULATION)
      class = 'N';
    else if (CTX->source == DSS_CORPUS)
      class = 'C';
     
    snprintf(x, sizeof(x), "%ld\t%c\t%s\t%s", 
            time(NULL), 
            class,
            (from == NULL) ? "<None Specified>" : from,
            (subject == NULL) ? "<None Specified>" : subject);
    for(y=0;y<strlen(x);y++)
      if (x[y] == '\n') 
        x[y] = 32;

#ifdef USER_LOGGING
    file = fopen(filename, "a");
    if (file != NULL) {
      int i = get_fcntl_lock(fileno(file));
      if (!i) {
          fputs(x, file);
          fputs("\n", file);
          free_fcntl_lock(fileno(file));
      } else {
        LOGDEBUG("Failed to lock %s: %d: %s\n", filename, i, strerror(errno));
      }

      fclose(file);
    }
#endif


#ifdef SYSTEM_LOGGING
    gettimeofday(&tv_end, NULL);

    snprintf(filename, sizeof(filename), "%s/system.log", DSPAM_HOME);
   
    file = fopen(filename, "a");
    if (file != NULL) {
      int i = get_fcntl_lock(fileno(file));
      if (!i) {
        double start, stop;
        char s[1024];

        start = tv_start.tv_usec/1000000.0;
        stop = tv_end.tv_usec/1000000.0;
        start += tv_start.tv_sec+0.0;
        stop += tv_end.tv_sec+0.0; 
        
        snprintf(s, sizeof(s), "%s\t%f\n", x, stop-start);
        fputs(s, file);
        free_fcntl_lock(fileno(file));
      } else {
        LOGDEBUG("Failed to lock %s: %d: %s\n", filename, i, strerror(errno));
      }
      fclose(file);
    }
#endif
  }
#endif

  /* FP and SM can return */
  if (CTX->message == NULL)
  {
    dspam_destroy (CTX);
    return result;
  }

#ifndef WEBMAIL
  /* Strip old X-DSPAM headers */
  node_nt = c_nt_first (CTX->message->components, &c_nt);
  if (node_nt != NULL && CTX->operating_mode == DSM_PROCESS)
  {
    struct _ds_message_block *block;

    while (node_nt != NULL)
    {
      block = node_nt->ptr;
      if (block->headers != NULL)
      {
        struct _ds_header_field *head;
        struct nt_node *node_header, *prev_node = block->headers->first, *old;

        node_header = block->headers->first;
        while(node_header != NULL) {
          head = (struct _ds_header_field *) node_header->ptr;

          if (result == DSR_ISSPAM &&
              head->heading        && 
              !strcasecmp(head->heading, "Subject")) {
            int do_tag = 0;
            char *subject;

#ifdef SPAM_SUBJECT
            do_tag = 1;
#endif
            if (PTX != NULL               && 
                PTX->spam_mode == DPS_TAG && 
                PTX->spam_subject[0] != 0)
              do_tag = 1; 

            if (do_tag) {
                char spam_subject[16];
                long subject_length;
                strcpy(spam_subject, "[SPAM]");
                if (PTX != NULL && PTX->spam_mode == DPS_TAG)
                  strcpy(spam_subject, PTX->spam_subject);

              subject_length = strlen(head->data)+strlen(spam_subject)+1;
              subject = malloc(subject_length);
              if (subject != NULL) {
                snprintf(subject, 
                         subject_length, "%s %s", 
                         spam_subject, 
                         head->data);
                free(head->data);
                head->data = subject;
              }
            }

          }
             
          if (head->heading && !strncmp(head->heading, "X-DSPAM", 7)
              && ((strcmp(head->heading, "X-DSPAM-Signature") ||
                 (session[0] != 0 || ATX->source != DSS_ERROR)))) {
            old = node_header;
            free(head->heading);
            free(head->data);

            if (node_header == block->headers->first) { 
              block->headers->first = node_header->next;
              prev_node = node_header->next;
              node_header = prev_node;
            } else {
              prev_node->next = node_header->next;
              node_header = node_header->next;
            }
            free(old->ptr);
            free(old);
            block->headers->items--;
            block->headers->insert = NULL;
          } else {
            prev_node = node_header;
            node_header = node_header->next;
          }
        }
      }
      node_nt = c_nt_next (CTX->message->components, &c_nt);
    }
  }

  /* Add our own X-DSPAM Headers */
  node_nt = c_nt_first (CTX->message->components, &c_nt);
  if (node_nt != NULL && ! FALSE_POSITIVE(CTX))
  {
    struct _ds_message_block *block = node_nt->ptr;
    if (block != NULL && block->headers != NULL)
    {
      struct _ds_header_field *head; 
      char data[128];

      snprintf(data, sizeof(data), "X-DSPAM-Result: %s", 
        (result == DSR_ISSPAM) ? "Spam" : "Innocent");

      head = _ds_create_header_field(data);
      if (head != NULL)
      {
#ifdef VERBOSE
        LOGDEBUG ("appending header %s: %s", head->heading, head->data)
#endif
          nt_add (block->headers, (void *) head);
      }
      else
        LOG (LOG_CRIT, ERROR_MEM_ALLOC);

      snprintf(data, sizeof(data), "X-DSPAM-Confidence: %01.4f", CTX->confidence);
      head = _ds_create_header_field(data);
      if (head != NULL)
      {
#ifdef VERBOSE
        LOGDEBUG("appending header %s: %s", head->heading, head->data);
#endif
        nt_add(block->headers, (void *) head);
      }
      else
        LOG (LOG_CRIT, ERROR_MEM_ALLOC);

      snprintf(data, sizeof(data), "X-DSPAM-Probability: %01.4f", CTX->probability);

      head = _ds_create_header_field(data);
      if (head != NULL)
      {
#ifdef VERBOSE
        LOGDEBUG ("appending header %s: %s", head->heading, head->data)
#endif
          nt_add (block->headers, (void *) head);
      }
      else
        LOG (LOG_CRIT, ERROR_MEM_ALLOC);

      if (CTX->training_mode != DST_NOTRAIN && session[0] != 0) {
#ifdef SIGNATURE_IN_HEADERS
        snprintf(data, sizeof(data), "X-DSPAM-Signature: %s%s%s", 
          SIGNATURE_BEGIN, session, SIGNATURE_END);
#else
        snprintf(data, sizeof(data), "X-DSPAM-Signature: %s", session);
#endif

        head = _ds_create_header_field(data);
        if (head != NULL)
        {
          if (session == NULL || strlen(session)<5) {
            LOGDEBUG("WARNING: Signature not generated, or invalid");
          }
#ifdef VERBOSE
          LOGDEBUG ("appending header %s: %s", head->heading, head->data)
#endif
            nt_add (block->headers, (void *) head);
        }
        else
          LOG (LOG_CRIT, ERROR_MEM_ALLOC);
      }

      if (result == DSR_ISSPAM)
      {

        snprintf(data, sizeof(data), "X-DSPAM-User: %s", username);
        head = _ds_create_header_field(data);
        if (head != NULL)
        {
#ifdef VERBOSE
          LOGDEBUG ("appending header %s: %s", head->heading, head->data)
#endif
            nt_add (block->headers, (void *) head);
        }
        else
          LOG (LOG_CRIT, ERROR_MEM_ALLOC);
      }
    }
  }
#endif /* !WEBMAIL */


#ifndef SIGNATURE_IN_ATTACHMENTS
#ifndef SIGNATURE_IN_HEADERS
#ifndef WEBMAIL
  /* Embed the signature into all text segments.
   * If for some reason `session' is empty string ("") then nothing
   * will be inserted.
   */
  i = 0;
  while (CTX->training_mode != DST_NOTRAIN && session[0] && node_nt != NULL)
  {
    struct _ds_message_block *block = node_nt->ptr;
    char *body_close = NULL, *dup = NULL;

    /* Append signature to subject of multipart/signed messages */
    if (block != NULL && block->media_type == MT_MULTIPART
        && block->media_subtype == MST_SIGNED && !i && block->boundary != NULL)
    {
      char boundary[strlen (block->boundary) + 1];
      char term[sizeof (boundary) + 2];
      struct nt_node *node_nt, *prev_node = NULL;
      struct nt_c c_nt;
      int i = 0;

      /* 
       * Message is already in an acceptable multipart format;
       * insert signature block
       */

      strlcpy (boundary, block->boundary, sizeof (boundary));
      snprintf (term, sizeof (term), "%s--", boundary);

      /* Strip the terminating boundary from the last block */
      node_nt = c_nt_first (CTX->message->components, &c_nt);
      while (node_nt != NULL)
      {

        block = (struct _ds_message_block *) node_nt->ptr;
        if (block->terminating_boundary != NULL
            && !strcmp (block->terminating_boundary, term) && term[0] != 0)
        {
          struct nt_node *old;
          free (block->terminating_boundary);
          block->terminating_boundary = strdup (boundary);
          term[0] = 0;

          old = node_nt->next;
          _ds_destroy_block ((struct _ds_message_block *) old->ptr);
          node_nt->next = old->next;
          free (old);
          CTX->message->components->insert = NULL;
          CTX->message->components->items--;
          break;
        }
        i++;
        prev_node = node_nt;
        node_nt = c_nt_next (CTX->message->components, &c_nt);
      }
      node_nt = c_nt_first (CTX->message->components, &c_nt);
      block = (struct _ds_message_block *) node_nt->ptr;

      if (boundary[0] != 0)
      {
        struct _ds_message_block *newblock;
        struct _ds_header_field *field;
        char scratch[128], typedata[128], term[512];

        snprintf (scratch, sizeof (scratch),
                  "%s%s%s\n", SIGNATURE_BEGIN, session, SIGNATURE_END);

        /* Create new message block */
        newblock =
          (struct _ds_message_block *)
          malloc (sizeof (struct _ds_message_block));
        if (newblock == NULL)
        {
          LOG (LOG_CRIT, ERROR_MEM_ALLOC);
          return EUNKNOWN;
        }

        newblock->headers = nt_create (NT_PTR);
        if (newblock->headers == NULL)
        {
          free (newblock);
          LOG (LOG_CRIT, ERROR_MEM_ALLOC);
          return EUNKNOWN;
        }

        /* Create new block information */
        snprintf (term, sizeof (term), "%s--\n\n", boundary);
        newblock->boundary = NULL;
        newblock->terminating_boundary = strdup (term);
        newblock->encoding = EN_7BIT;
        newblock->original_encoding = EN_7BIT;
        newblock->media_type = MT_TEXT;
        newblock->media_subtype = MST_PLAIN;
        newblock->body = buffer_create (scratch);
        newblock->original_signed_body = NULL;

        snprintf(typedata, sizeof(typedata),
            "Content-Type: text/plain; name=\"dspam.txt\"");
        field = _ds_create_header_field(typedata);
        if (field == NULL)
        {
          _ds_destroy_block (newblock);
          LOG (LOG_CRIT, ERROR_MEM_ALLOC);
          return EUNKNOWN;
        }
        nt_add (newblock->headers, field);

        snprintf(typedata, sizeof(typedata),
            "Content-Disposition: attachment");
        field = _ds_create_header_field(typedata);
        if (field == NULL)
        {
          _ds_destroy_block (newblock);
          LOG (LOG_CRIT, ERROR_MEM_ALLOC);
          return EUNKNOWN;
        }
        nt_add (newblock->headers, field);


        snprintf(typedata, sizeof(typedata), 
            "Content-Transfer-Encoding: 7bit"); 
        field = _ds_create_header_field(typedata);
        if (field == NULL)
        {
          _ds_destroy_block (newblock);
          LOG (LOG_CRIT, ERROR_MEM_ALLOC);
          return EUNKNOWN;
        }
        nt_add (newblock->headers, field);

        /* Add the new block */

        nt_add (CTX->message->components, (void *) newblock);
      }

      break;
    }

    if (block != NULL
        && (block->media_type == MT_TEXT
            || (block->boundary == NULL && i == 0
                && block->media_type != MT_MULTIPART)))
    {
      int unclosed_html = 0, is_attachment = 0;
      struct _ds_header_field *field;
      struct nt_node *node_hnt;
      struct nt_c c_hnt;

      node_hnt = c_nt_first (block->headers, &c_hnt);
      while (node_hnt != NULL)
      {
        field = (struct _ds_header_field *) node_hnt->ptr;
        if (field != NULL && field->heading != NULL && field->data != NULL)
          if (!strncasecmp (field->heading, "Content-Disposition", 19))
            if (!strncasecmp (field->data, "attachment", 10))
              is_attachment = 1;
        node_hnt = c_nt_next (block->headers, &c_hnt);
      }

      if (is_attachment)
      {
        node_nt = c_nt_next (CTX->message->components, &c_nt);
        i++;
        continue;
      }

      if (block->body != NULL && block->body->data != NULL
          && block->media_subtype == MST_HTML)
        body_close = strstr (block->body->data, "</body");
      if (body_close == NULL && block->body != NULL
          && block->body->data != NULL && block->media_subtype == MST_HTML)
        body_close = strstr (block->body->data, "</BODY");
      if (body_close == NULL && block->body != NULL
          && block->body->data != NULL && block->media_subtype == MST_HTML)
        body_close = strstr (block->body->data, "</Body");

      if (body_close != NULL)
      {
        dup = strdup (body_close);
        block->body->used -= strlen (dup);
        body_close[0] = 0;
      }

      if (block->body->data != NULL
          &&
          ((strstr (block->body->data, "</html")
            || strstr (block->body->data, "</HTML"))
           && (!strstr (block->body->data, "</html>")
               && !strstr (block->body->data, "</HTML>"))))
        unclosed_html = 1;

      if (!dup && unclosed_html)
        buffer_cat (block->body, ">");

      buffer_cat (block->body, "\n");
      buffer_cat (block->body, SIGNATURE_BEGIN);
      buffer_cat (block->body, session);
      buffer_cat (block->body, SIGNATURE_END);
      buffer_cat (block->body, "\n\n");

      if (dup)
      {
        buffer_cat (block->body, dup);
        free (dup);
        if (unclosed_html)
          buffer_cat (block->body, ">");
      }
    }
    node_nt = c_nt_next (CTX->message->components, &c_nt);
    i++;
  }
#endif /* !WEBMAIL */
#endif /* !SIGNATURE_IN_HEADERS */
#endif /* !SIGNATURE_IN_ATTACHMENTS */

  /* reconstruct message from components */

  copyback = _ds_assemble_message (CTX->message);
  buffer_clear (message);
  buffer_cat (message, copyback);
  free (copyback);

#ifdef SOURCE_ADDRESS_TRACKING
  if (
       (
       (CTX->training_mode == DST_TOE && CTX->operating_mode == DSM_CLASSIFY) ||
       (CTX->training_mode != DST_TOE && CTX->operating_mode == DSM_PROCESS)
       ) &&
       CTX->source != DSS_CORPUS)
  {
    char ip[32];
    char mta_whitelist[MAX_FILENAME_LENGTH];
    snprintf (mta_whitelist, MAX_FILENAME_LENGTH,
              "%s/mta.whitelist", DSPAM_HOME);

    if (!dspam_getsource (CTX, ip, sizeof (ip), mta_whitelist))
    {
      if (CTX->totals.innocent_learned + CTX->totals.innocent_classified 
          > 2500) {
        if (result == DSR_ISSPAM) {
          FILE *file;
          char dropfile[MAX_FILENAME_LENGTH];
          LOG (LOG_INFO, "spam detected from %s", ip);
          snprintf(dropfile, sizeof(dropfile), "/var/spool/sbl/%s", ip);
          file = fopen(dropfile, "w");
          if (file != NULL) 
            fclose(file);
        }
        else {
          LOG (LOG_DEBUG, "innocent message from %s", ip);
        }
      }
    }
  }
#endif

  if (CTX->operating_mode == DSM_CLASSIFY)
  {
    printf("X-DSPAM-Result: %s; result=\"%s\"; probability=%01.4f; "
           "confidence=%02.2f\n",
           CTX->username,
           (result == DSR_ISSPAM) ? "Spam" : "Innocent",
           CTX->probability,
           CTX->confidence);
  }

  if (CTX->message != NULL)
    _ds_destroy_message (CTX->message);
  dspam_destroy (CTX);

  nt_destroy (inoc_users);
  nt_destroy (classify_users);

  return result;
}

int
deliver_message (const char *message, const char *mailer_args,
                 const char *username)
{
#ifdef DELIVER_TO_STDOUT
  fputs (message, stdout);
#else
  char args[1024];
  char *margs, *mmargs, *arg;
  FILE *file;
  int rc;

  if (mailer_args == NULL)
  {
    fputs (message, stdout);
    return 0;
  }

  args[0] = 0;
  margs = strdup (mailer_args);
  mmargs = margs;
  arg = strsep (&margs, " ");
  while (arg != NULL)
  {
    char a[256], b[256];
    int i;
 
    if (!strcmp (arg, "$u") || !strcmp (arg, "\\$u") || !strcmp (arg, "%u"))
      strlcpy(a, username, sizeof(a));
    else
      strlcpy(a, arg, sizeof(a));

    for(i=0;i<strlen(a);i++) {
      if (!(isalnum((int) a[i]) || a[i] == '+' || a[i] == '_' || a[i] == '-' ||
                 a[i] == '.' || a[i] == '/' || a[i] == '@')) {
        strlcpy(b, a+i, sizeof(b));
        a[i] = '\\';
        a[i+1] = 0;
        strlcat(a, b, sizeof(a));
        i++;
      }
    }

    if (arg != NULL) {
      strlcat (args, a, sizeof(args));
    }
    arg = strsep(&margs, " ");

    if (arg) {
      strlcat (args, " ", sizeof (args));
    }
  }
  free (mmargs);

  LOGDEBUG ("Opening pipe to LDA: %s", args);
  file = popen (args, "w");

  if (file == NULL)
  {
    file_error ("Error opening pipe to LDA", args, strerror (errno));
    return EFILE;
  }

  fputs (message, file);
  rc = pclose (file);
  if (rc == -1)
  {
    file_error ("Error obtaining exit status of LDA", args, strerror (errno));
    return EFILE;
  }
  else if (WIFEXITED (rc))
  {
    int lda_exit_code;
    lda_exit_code = WEXITSTATUS (rc);
    if (lda_exit_code == 0)
    {
      LOGDEBUG ("LDA returned success");
    }
    else
    {
      report_error_printf
        ("LDA returned error, exit code: %d, LDA command line: %s",
         lda_exit_code, args);
      return lda_exit_code;
    }
  }
  else if (WIFSIGNALED (rc))
  {
    int sig;
    sig = WTERMSIG (rc);
    report_error_printf
      ("LDA terminated by signal #%d, LDA command line: %s", sig, args);
    return sig;
  }
  else
  {
    report_error_printf ("Unexpected ret-code from pclose(): %d\n", rc);
    return rc;
  }
#endif

  return 0;
}

int
quarantine_message (const char *message, const char *username)
{
  char filename[MAX_FILENAME_LENGTH];
  FILE *file;
  int line = 1;
  char *x, *msg = strdup (message);
  int i;

  if (msg == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    return EUNKNOWN;
  }

  strcpy (filename, _ds_userdir_path (username, "mbox"));
  file = fopen (filename, "a");
  if (file == NULL)
  {
    file_error (ERROR_FILE_WRITE, filename, strerror (errno));
    free (msg);
    return EFILE;
  }

  i = get_fcntl_lock(fileno(file));
  if (i) {
    LOGDEBUG("Failed to lock %s: Error %d: %s\n", filename, i, strerror(errno));
    free(msg);
    return EFILE;
  }

  /* From header */
  if (strncmp (message, "From ", 5))
  {
    char head[128];
    time_t tm = time (NULL);

    snprintf (head, sizeof (head), "From QUARANTINE %s", ctime (&tm));
    fputs (head, file);
  }

  x = strsep (&msg, "\n");
  while (x != NULL)
  {
    if (!strncmp (x, "From ", 5) && line != 1)
      fputs (">", file);
    fputs (x, file);
    fputs ("\n", file);
    line++;
    x = strsep (&msg, "\n");
  }
  fputs ("\n\n", file);

  free_fcntl_lock(fileno(file));
  fclose (file);

  free (msg);
  return 0;
}

int
write_web_stats (const char *username, struct _ds_spam_totals *totals)
{
  char filename[MAX_FILENAME_LENGTH];
  FILE *file;

  if (totals == NULL)
  {
    LOGDEBUG ("totals are null\n");
    return EINVAL;
  }

  strcpy (filename, _ds_userdir_path (username, "stats"));
  _ds_prepare_path_for (filename);
  file = fopen (filename, "w");
  if (file == NULL)
  {
    file_error (ERROR_FILE_WRITE, filename, strerror (errno));
    return EFILE;
  }

  fprintf (file, "%ld,%ld,%ld,%ld,%ld,%ld\n",
           (totals->spam_learned + totals->spam_classified) - 
             (totals->spam_misclassified + totals->spam_corpusfed),
           (totals->innocent_learned + totals->innocent_classified) -
             (totals->innocent_misclassified + totals->innocent_corpusfed),
           totals->spam_misclassified, totals->innocent_misclassified,
           totals->spam_corpusfed, totals->innocent_corpusfed);
  fclose (file);
  return 0;
}

/* function: inoculate_user
   parameters: only SIG _OR_ message should be passed.  the other
               should be left NULL depending on whether we have
               a signature or a message
*/

int
inoculate_user (const char *username, struct _ds_spam_signature *SIG,
                const char *message)
{
  DSPAM_CTX *INOC;
  int do_inoc = 1, result = 0;
  int f_all = 0;

  LOGDEBUG ("checking to see if user %s requires this inoculation", username);

  /* First see if the user needs to be inoculated */
#ifdef CHAINED_TOKENS
  f_all = f_all ^ DSF_CHAINED;
#endif

#ifdef BNR
  f_all = f_all ^ DSF_NOISE;
#endif

  INOC = dspam_init (username, NULL, DSM_CLASSIFY, f_all);
  if (INOC)
  {
    if (SIG)
    {
      INOC->flags |= DSF_SIGNATURE;
      INOC->signature = SIG;
      result = dspam_process (INOC, NULL);
    }
    else
    {
      INOC->signature = NULL;
      result = dspam_process (INOC, message);
    }

    if (!result && INOC->result == DSR_ISSPAM)
      do_inoc = 0;

    if (SIG)
      INOC->signature = NULL;
    dspam_destroy (INOC);
  }

  if (!do_inoc)
  {
    LOGDEBUG ("skipping user %s: doesn't require inoculation", username);
    return EFAILURE;
  }
  else
  {
    LOGDEBUG ("inoculating user %s", username);

#ifdef CHAINED_TOKENS
  f_all = f_all ^ DSF_CHAINED;
#endif
                                                                                
#ifdef BNR
  f_all = f_all ^ DSF_NOISE;
#endif
    INOC = dspam_init (username, NULL, DSM_PROCESS, f_all);
    if (INOC)
    {
      INOC->classification = DSR_ISSPAM;
      INOC->source = DSS_INOCULATION;
      if (SIG)
      {
        INOC->flags |= DSF_SIGNATURE;
        INOC->signature = SIG;
        result = dspam_process (INOC, NULL);
      }
      else
      {
        result = dspam_process (INOC, message);
      }

      if (SIG)
        INOC->signature = NULL;
      dspam_destroy (INOC);
    }
  }

  return result;
}

/* user_classify: classifies message for another user
   parameters: only SIG _OR_ message should be passed.  the other
               should be left NULL depending on whether we have
               a signature or a message
*/


int
user_classify (const char *username, struct _ds_spam_signature *SIG,
               const char *message)
{
  DSPAM_CTX *CLX;
  int result = 0;
  int f_all = 0;

#ifdef CHAINED_TOKENS
  f_all = f_all ^ DSF_CHAINED;
#endif
                                                                                
#ifdef BNR
  f_all = f_all ^ DSF_NOISE;
#endif
                                                                                
  /* First see if the user needs to be inoculated */
  CLX = dspam_init (username, NULL, DSM_CLASSIFY, f_all);
  if (CLX)
  {
    if (SIG)
    {
      CLX->flags |= DSF_SIGNATURE;
      CLX->signature = SIG;
      result = dspam_process (CLX, NULL);
    }
    else
    {
      result = dspam_process (CLX, message);
    }

    if (SIG)
      CLX->signature = NULL;

    if (result)
    {
      LOGDEBUG ("user_classify() returned error %d", result);
      result = 0;
    }
    else
      result = CLX->result;

    dspam_destroy (CLX);
  }

  return result;
}

int send_notice(const char *filename, const char *mailer_args, 
                const char *username) {
  FILE *f;
  char msgfile[MAX_FILENAME_LENGTH];
  buffer *b;
  char buff[1024];
                                                                                
  snprintf(msgfile, sizeof(msgfile), "%s/txt/%s", DSPAM_HOME, filename);
  f = fopen(msgfile, "r");
  if (f == NULL) 
    return EFILE;

  b = buffer_create(NULL);
  if (b == NULL) {
    report_error(ERROR_MEM_ALLOC);
    return EUNKNOWN;
  }

  while(fgets(buff, sizeof(buff), f)!=NULL) {
    char *s = buff;
    char *w = strstr(buff, "$u");
    while(w != NULL) {
      w[0] = 0;
      buffer_cat(b, s);
      buffer_cat(b, username);
        s = w+2;
        w = strstr(s, "$u");
    }
    buffer_cat(b, s);
  }
  fclose(f);
  deliver_message(b->data, mailer_args, username);

  buffer_destroy(b);

  return 0;
}


#ifdef TRUSTED_USER_SECURITY
int is_trusted(struct passwd *p) {
  FILE *file;
  char filename[MAX_FILENAME_LENGTH];
  char buff[1024];
  int trusted = 0;

  snprintf (filename, MAX_FILENAME_LENGTH, "%s/trusted.users", DSPAM_HOME);
  file = fopen (filename, "r");
  if (file != NULL)
  {
    while (fgets (buff, sizeof (buff), file) != NULL)
    {
      ALLTRIM(buff);
      if (!strcmp (buff, p->pw_name))
        trusted = 1;
    }
    fclose (file);
  }
  else
  {
    LOG (LOG_CRIT,
         "unable to open %s/trusted.users for reading: %s.",
         DSPAM_HOME, strerror (errno));
  }
                                                                                
  if (!trusted)
  {
    LOG (LOG_INFO, "forcing username for untrusted user %s", p->pw_name);
  }

  return trusted;
}
#endif

PREF *load_preferences(const char *username) {
  char filename[MAX_FILENAME_LENGTH];
  PREF *PTX = malloc(sizeof(PREF));
  char buff[128];
  FILE *file;
  char *p, *q;

  if (PTX == NULL)
    return NULL;

  PTX->training_mode = -1;
  PTX->spam_mode = -1;
  PTX->spam_subject[0] = 0;
  PTX->feature_bnr = -1;
  PTX->feature_whitelist = -1;
  PTX->training_buffer = -1;

  if (username == NULL) {
    snprintf(filename, MAX_FILENAME_LENGTH, "%s/default.prefs", DSPAM_HOME);
  } else {
    strcpy (filename, _ds_userdir_path (username, "prefs"));
  }
  file = fopen(filename, "r");
  if (file == NULL) { 
    free(PTX);
    return (username == NULL) ? NULL : load_preferences(NULL); 
  }
 
  while(fgets(buff, sizeof(buff), file)!=NULL) {
    p = strtok(buff, "=");
    if (p != NULL)
      q = strtok(NULL, "=");
    if (p != NULL && q != NULL) {
      chomp(q);
      if (!strcmp(p, "trainingMode")) {
        if (!strcmp(q, "TOE")) 
          PTX->training_mode = DST_TOE;
        else if (!strcmp(q, "TEFT"))
          PTX->training_mode = DST_TEFT;
        else if (!strcmp(q, "TUM"))
          PTX->training_mode = DST_TUM;
      } else if (!strcmp(p, "spamAction")) {
        if (!strcmp(q, "quarantine")) 
          PTX->spam_mode = DPS_QUARANTINE; 
        else if (!strcmp(q, "tag"))
          PTX->spam_mode = DPS_TAG;
        else if (!strcmp(q, "deliver"))
          PTX->spam_mode = DPS_DELIVER;
      } else if (!strcmp(p, "spamSubject")) {
        strlcpy(PTX->spam_subject, q, sizeof(PTX->spam_subject));
      } else if (!strcmp(p, "enableBNR")) {
        if (!strcmp(q, "on")) 
          PTX->feature_bnr = 1;
        else
          PTX->feature_bnr = 0;
      } else if (!strcmp(p, "enableWhitelist")) {
        if (!strcmp(q, "on"))
          PTX->feature_whitelist = 1;
        else
          PTX->feature_whitelist = 0; 
      } else if (!strcmp(p, "statisticalSedation") && q[0] != 0) {
        PTX->training_buffer = atoi(q);
      }
    }
  }
  fclose(file);

  return PTX;
}

