/*
 DSPAM
 COPYRIGHT (C) 2003 NETWORK DWEEBS CORPORATION

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifdef HAVE_CONFIG_H
#include <auto-config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>

#ifdef TIME_WITH_SYS_TIME
#   include <sys/time.h>
#   include <time.h>
#else
#   ifdef HAVE_SYS_TIME_H
#       include <sys/time.h>
#   else
#       include <time.h>
#   endif
#endif

#include "error.h"
#include "util.h"
#include "config.h"

/*  Subroutine: report_error
   Description: Reports an error to the supported facilities (currently stderr)
    Parameters: Error message to report
*/

void
report_error (const char *error)
{
  time_t tm;
  char ct[128];

#ifdef DEBUG
  debug (error);
#endif

  tm = time (NULL);
  snprintf (ct, sizeof (ct), "%s", ctime (&tm));
  chomp (ct);

  fprintf (stderr, "%ld: [%s] %s\n", (long) getpid (), ct, error);

  return;
}

void
report_error_printf (const char *fmt, ...)
{
  va_list ap;
  char buff[1024];
  va_start (ap, fmt);
  vsnprintf (buff, sizeof (buff), fmt, ap);
  report_error (buff);
  va_end (ap);
  return;
}

void
file_error (const char *operation, const char *filename, const char *error)
{
  char errmsg[1024];

  snprintf (errmsg, sizeof (errmsg), "%s: %s: %s", operation, filename,
            error);
  report_error (errmsg);
  return;
}

#ifdef DEBUG
void
debug (const char *text)
{
  FILE *file;
  time_t tm;
  char ct[128];
  char fn[MAX_FILENAME_LENGTH];

  snprintf (fn, sizeof (fn), "%s/dspam.debug", DSPAM_HOME);

  tm = time (NULL);
  snprintf (ct, sizeof (ct), "%s", ctime (&tm));
  chomp (ct);

  /* TODO: importatnt: checking for "is opened file a symlink?" !!! */
  file = fopen (fn, "a");
  if (file != NULL)
  {
    fprintf (file, "%ld: [%s] %s\n", (long) getpid (), ct, text);
    fclose (file);
  }
  return;
}
#endif

#ifndef HAVE_ISO_VARARGS
void
LOGDEBUG (char *text, ...)
{
#ifdef DEBUG
  char debug_text[1024];
  va_list args;

  if (!DO_DEBUG)
    return;

  va_start (args, text);
  snprintf (debug_text, sizeof (debug_text), text, args);
  va_end (args);

  debug (debug_text);
#endif
}

void
LOG (int priority, char *text, ...)
{
#ifdef DEBUG
  va_list args;

  va_start (args, text);
  openlog ("dspam", LOG_PID, LOG_MAIL);
  syslog (priority, text, args);
  closelog ();
  LOGDEBUG (text, args);
  va_end (args);
#endif
}
#endif
